from ruly_dmn.common import (ModelHandler,
                             RuleFactory)
from ruly_dmn.dmn import (DMN,
                          rule_factory_cb)
from ruly_dmn.handlers.camunda_modeler import (CamundaModelerHandler)


__all__ = ['DMN',
           'ModelHandler',
           'RuleFactory',
           'rule_factory_cb',
           'CamundaModelerHandler']



import argparse
import json
import sys
from pathlib import Path

import ruly_dmn.dmn
from ruly_dmn.handlers.camunda_modeler import CamundaModelerHandler


def main():
    args = _create_parser().parse_args()
    handler = CamundaModelerHandler(args.file, args.output_file)
    dmn = ruly_dmn.dmn.DMN(handler)
    inputs = {}
    for k, v in (a.split('=', 2) for a in args.inputs):
        try:
            inputs[k] = json.loads(v)
        except json.JSONDecodeError:
            inputs[k] = v
    print(args.goal, '=', dmn.decide(inputs, args.goal))


def _create_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('file', type=Path,
                        help='DMN file.')
    parser.add_argument('goal',
                        help='Name of the goal decision.')
    parser.add_argument('inputs', nargs='*',
                        help='Input value pairs in format '
                        '<input name>=<value>, where value will be attempted '
                        'to be parsed as JSON, otherwise it will be '
                        'interpreted as a string. '
                        'i.e. a=1 b=xyz c=\'"{"json": "value"}"\'')
    parser.add_argument('--output-file', '-o', metavar='path', type=Path,
                        help='Output DMN file path, in case new rules are '
                        'added. Can be the same path as the input file. If '
                        'not set, DMN changes are not written anywhere.',
                        default=None)

    return parser


if __name__ == '__main__':
    sys.exit(main())



import abc


class ModelHandler(abc.ABC):
    """Abstract class representing a connection between an encoded DMN model
    and ruly's rules. Implementation should parse initial rules from the model
    and be available for updates with new rules."""

    @abc.abstractproperty
    @property
    def dependencies(self):
        """Dict[Tuple(str), Tuple(str)]: output-input variable name pairs"""

    @abc.abstractproperty
    @property
    def rules(self):
        """List[ruly.Rule]: all rules available in the model"""
        return self._rules

    @abc.abstractmethod
    def update(self, knowledge_base):
        """Updates handler with a new knowledge base

        Args:
            knowledge_base (ruly.KnowledgeBase): new knowledge base
        """


class RuleFactory(abc.ABC):
    """Abstract class whose instances should implement rule creation methods"""

    @abc.abstractmethod
    def create_rule(self, state, fired_rules, output_names):
        """Creates a new rule

        Args:
            state (Dict[str, Any]): state
            fired_rules (List[ruly.Rule]): rules activated for state and output
                combination
            output_names (Tuple[str]): output values for the new rule's
                consequent

        Returns:
            Optional[ruly.Rule]: generated rule, if None, new rule couldn't be
            created with given inputs"""



from frozendict import frozendict
import itertools
import json
import ruly

from ruly_dmn import common


class DMN:
    """Class that contains the DMN implementation.

    Args:
        handler (ruly_dmn.ModelHandler): model handler
        rule_factory_cb (Optional[Callable]): function that creates a rule
            factory - if None, a factory that uses console is used. Signature
            should match the signature of :func:`ruly_dmn.rule_factory_cb`"""

    def __init__(self, handler, rule_factory_cb=None):
        self._handler = handler
        self._knowledge_base = ruly.KnowledgeBase(*handler.rules)
        self._factory_cb = rule_factory_cb

        all_outputs = set(itertools.chain(*handler.dependencies.keys()))
        all_inputs = set(itertools.chain(*handler.dependencies.values()))
        self._inputs = all_inputs - all_outputs

    @property
    def inputs(self):
        """List[str]: input variables for all available decisions"""
        return self._inputs

    def decide(self, inputs, decision):
        """Attempts to solve for decision based on given inputs. May create
        new rules if the factory creates them.

        Args:
            inputs (Dict[str, Any]): name-value pairs of all inputs
            decision (str): name of the decision that should be resolved

        Returns:
            Any: calculated decision"""
        rules = list(self._knowledge_base.rules)
        if self._factory_cb is None:
            rule_factory = _ConsoleRuleFactory(self._handler)
        else:
            rule_factory = self._factory_cb(self._handler)

        def post_eval_cb(state, output_name, fired_rules):
            new_rule = rule_factory.create_rule(state, fired_rules,
                                                output_name)
            if new_rule is not None and new_rule not in rules:
                if len(fired_rules) == 0:
                    rules.append(new_rule)
                else:
                    rules.insert(rules.index(fired_rules[0]), new_rule)
                raise _CancelEvaluationException()
            elif len(fired_rules) > 0:
                state = dict(state, **fired_rules[0].consequent)
            return state

        state = None
        rule_count = len(rules)
        rules_changed = False
        while state is None:
            try:
                state = ruly.backward_chain(self._knowledge_base, decision,
                                            post_eval_cb=post_eval_cb,
                                            **inputs)
            except _CancelEvaluationException:
                if len(rules) == rule_count:
                    break
                else:
                    rules_changed = True
                    self._knowledge_base = ruly.KnowledgeBase(*rules)

        if rules_changed:
            self._handler.update(self._knowledge_base)

        return state[decision]


def rule_factory_cb(handler):
    """Placeholder function containing the signature for rule factory callbacks

    Args:
        handler (ruly_dmn.ModelHandler): model handler

    Returns:
        ruly_dmn.RuleFactory: rule factory"""


class _ConsoleRuleFactory(common.RuleFactory):

    def __init__(self, handler):
        self._rejections = []
        self._handler = handler

    def create_rule(self, state, fired_rules, output_name):
        output_names = [output_names for output_names
                        in self._handler.dependencies
                        if output_name in output_names][0]
        input_names = self._handler.dependencies[output_names]
        input_values = {name: state[name] for name in input_names
                        if state[name] is not None}
        if (input_values, output_names) in self._rejections:
            return None
        if not self._show_prompts(fired_rules, state, input_names):
            return None
        if len(fired_rules) == 0:
            question = (f'Unable to decide {output_names} for inputs '
                        f'{input_values}, generate new rule?')
        else:
            question = (f'Fired rules for {output_names} did not use all '
                        f'available input decisions - {input_values}, would '
                        f'you like to create a new rule that does?')
        if not self._confirm_creation(question):
            self._rejections.append((input_values, output_names))
            return None

        rule = self._create_prompt(input_values, output_names)
        print('Created rule:', rule)
        return rule

    def _show_prompts(self, fired_rules, state, input_names):
        available_inputs = {name for name in input_names
                            if state[name] is not None}
        for rule in fired_rules:
            rule_deps = set(ruly.get_rule_depending_variables(rule))
            if (rule_deps & available_inputs) == available_inputs:
                return False
        return True

    def _confirm_creation(self, question):
        answer = input(f'{question} (Y/n) ').lower() or 'y'
        while answer not in ('y', 'n'):
            answer = input('Please type y or n. (Y/n)').lower() or 'y'
        if answer == 'n':
            return False
        return True

    def _create_prompt(self, input_values, output_names):
        antecedent = ruly.Expression(
            ruly.Operator.AND, tuple(ruly.EqualsCondition(name, value)
                                     for name, value in input_values.items()
                                     if value is not None))
        print('Please type the expected output values (JSON)')
        print(f'IF {antecedent} THEN')
        assignments = {}
        for output_name in output_names:
            value = None
            while value is None:
                value_json = input(f'{output_name} = ')
                try:
                    value = json.loads(value_json)
                except json.JSONDecodeError:
                    print('Invalid JSON string, please try again')
            assignments[output_name] = value
        return ruly.Rule(antecedent, frozendict(assignments))


class _CancelEvaluationException(Exception):
    pass



import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()
with open('version.txt', 'r') as fh:
    version = fh.read()[:-1]

setuptools.setup(
    name='ruly-dmn',
    packages=['ruly_dmn'],
    version=version,
    url='https://github.com/ZlatSic/ruly-dmn',
    author='Zlatan Sičanica',
    author_email='zlatan.sicanica@gmail.com',
    description='DMN implementation with ruly rule engine',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities'
    ],
    entry_points={
        'console_scripts': [
            'ruly-dmn = ruly_dmn.main:main'
        ]
    },
    install_requires=[
        'ruly',
        'frozendict'
    ]
)



import pytest
import ruly

import ruly_dmn.common
import ruly_dmn.dmn


class MockModelHandler(ruly_dmn.common.ModelHandler):

    def __init__(self, dependencies={}, rules=[], update_fn=None):
        self._dependencies = dependencies
        self._rules = rules
        self._update_fn = update_fn

    @property
    def dependencies(self):
        return self._dependencies

    @property
    def rules(self):
        return self._rules

    def update(self, knowledge_base):
        if self._update_fn is not None:
            return self._update_fn(knowledge_base)


class MockRuleFactory(ruly_dmn.common.RuleFactory):

    def __init__(self, create_rule_fn=None):
        self._create_rule_fn = create_rule_fn

    def create_rule(self, state, fired_rules, output_names):
        if self._create_rule_fn is not None:
            return self._create_rule_fn(state, fired_rules, output_names)


def test_inputs():
    handler = MockModelHandler({('a', 'b'): ('c', 'd', 'e'),
                                ('e', ): ('f', 'g')})
    dmn = ruly_dmn.dmn.DMN(handler)
    assert dmn.inputs == {'c', 'd', 'f', 'g'}


@pytest.mark.parametrize('rules,inputs,decision,expected', [
    ([ruly.Rule(ruly.EqualsCondition('x', 1), {'y': 2})], {'x': 1}, 'y', 2),
    ([ruly.Rule(ruly.EqualsCondition('x', 1), {'y': 2}),
      ruly.Rule(ruly.EqualsCondition('y', 2), {'z': 3})], {'x': 1}, 'z', 3),
])
def test_decide(rules, inputs, decision, expected):
    handler = MockModelHandler(rules=rules)
    dmn = ruly_dmn.dmn.DMN(handler, lambda _: MockRuleFactory())
    assert dmn.decide(inputs, decision) == expected


def test_factory_called():
    called = False
    create_args = {}

    new_rule = ruly.Rule(ruly.EqualsCondition('x', 2), {'y': 3})

    def factory_cb(*args):
        nonlocal called
        called = True
        return MockRuleFactory(create_rule_fn)

    def create_rule_fn(state, fired_rules, output_names):
        create_args['state'] = state
        create_args['fired_rules'] = fired_rules
        create_args['output_names'] = output_names
        return new_rule

    rules = [ruly.Rule(ruly.EqualsCondition('x', 1), {'y': 2})]
    dmn = ruly_dmn.dmn.DMN(MockModelHandler(rules=rules), factory_cb)
    dmn.decide({'x': 1}, 'y')
    assert called
    assert create_args == {'state': {'x': 1, 'y': None},
                           'fired_rules': rules,
                           'output_names': 'y'}
    assert dmn._knowledge_base.rules == tuple([new_rule] + rules)
